-- ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
-- Written by Kimberly L. Tripp, SQLskills.com
-- All Rights Reserved.
-- 
-- For more scripts, sample code and additional training resources,
-- go to http://www.SQLskills.com
--	Immerse yourself in SQL Server
--
-- Disclaimer - Thoroughly test this script, execute at your own risk.
-- ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

USE Credit
go

-- Review the indexes on the charge table:
sp_helpindex charge
go

-- Drop any indexes that reference the charge_amt column
-- (NOTE: you might not have any depending on how you installed credit
--        and/or whether or not you've run other scripts.)
-- DROP INDEX Charge.charge_charge_Amt
-- DROP INDEX Charge.Covering1
-- DROP INDEX Charge.Covering2
go

-- Just to see how many rows there are s/b 100K or 1.6 million
SELECT count(*) 
FROM dbo.charge AS c
go

-- Also for reference, see the maximum charge amount
SELECT MAX(c.charge_amt) 
FROM dbo.charge AS c
go

-- See the cost (in terms of I/O of the reads for charges)
SET STATISTICS IO ON
go

SELECT c.* 
FROM charge AS c
WHERE c.charge_amt > 6000
go
-- 0 ROWS... And even though there are NO rows that match there are still many I/Os 
--           because a table scan was performed. SQL Server didn't know that there
--           were no rows - only we did! 

-- Add a constraint that enforces a maximum value. This doesn't really make sense
-- here as you probably don't want to limit charge amounts however, the impact (in
-- terms of performance and I/O) will be obvious!

ALTER TABLE dbo.charge
ADD CONSTRAINT ChargeChargeAmtCK
	CHECK (charge_amt < 5010)
go

-- Run the same query again
SELECT c.* 
FROM charge AS c
WHERE c.charge_amt > 6000
go
-- 0 ROWS... and 0 I/Os. SQL Server now knows that there are no rows. It KNOWS for
--           certainty because the constraint is trusted...

-- What if the constraint were not trusted... First, we need to create one that's not 
-- trusted. Drop the current version.
ALTER TABLE charge
	DROP CONSTRAINT ChargeChargeAmtCK
go

-- Add the constraint with NOCHECK. Adding NOCHECK tells SQL Server NOT to check the
-- data. If the data's not checked then the values are not guaranteed (constraint is not
-- trusted).
ALTER TABLE dbo.charge
	WITH NOCHECK -- This is evil :)
ADD CONSTRAINT ChargeChargeAmtCK
	CHECK (charge_amt < 5010)
go

SELECT c.* 
FROM charge AS c
WHERE c.charge_amt > 6000
go
-- 0 ROWS... BACK TO A SCAN...limit your use of WITH NOCHECK!

-- Is there a way to make it TRUSTED again... unfortunately NO.
-- Even if you use DBCC CHECKCONSTRAINTS and SQL Server doesn't find any invalid data, the
-- constraint won't be trusted. 

-- To verify the constraint settings:
SELECT OBJECTPROPERTY(object_id('ChargeChargeAmtCK'), 'CnstIsNOTTrusted') 
-- should be 1 because the constraint is not trusted...
go

-- To verify the constraint against the data
DBCC CHECKCONSTRAINTS('charge')
go

SELECT OBJECTPROPERTY(object_id('ChargeChargeAmtCK'), 'CnstIsNOTTrusted') 
go

-- Even if the data did not produce errors it's still NOT trusted because only the create 
-- with CHECK properly sets this bit! (It's because of the way that WITh CHECK locks the
-- table to verify the data; DBCC CHECKCONSTRAINTS does not lock the table the same way.)

-- IF you absolutely want the data consistency enforced by having a check constraint immeditately
-- (meaning 10am on a Monday) and you can't afford the table lock that create WITH CHECK uses 
-- then you can add the constraint with NOCHECK NOCHECK (at 10am) and then SCHEDULE an off hours 
-- DROP/re-CREATE of the constraint. 
-- This will ensure that the optimizer properly uses the constraint wherever it can! No matter what,
-- you'll have "downtime" of about the equivalent of a table scan for that table.